package eu.mhsl.minecraft.pixelblocks.pixelblock;

import eu.mhsl.minecraft.pixelblocks.chunkGenerators.EmptyChunkGenerator;
import org.bukkit.*;
import org.bukkit.block.data.BlockData;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Interaction;
import org.codehaus.plexus.util.FileUtils;

import java.io.File;
import java.io.IOException;
import java.sql.*;
import java.util.*;

import static eu.mhsl.minecraft.pixelblocks.PixelBlocks.plugin;

public class PixelBlock {
    public static List<PixelBlock> placedBlocks = new ArrayList<>();
    public static float hitboxOffset = 0.005F;
    public static int worldGrassBorderWidth = 5;

    public Location pixelBlockLocation;
    public double pixelsPerBlock;
    public ArrayList<Pixel> pixels = new ArrayList<>();
    public Interaction hitbox;

    public long lastEntryTime;
    public Location lastEntryLocation;
    public UUID owner;
    public UUID uuid;

    public PixelBlock(Location originLocation, UUID owner, double pixelsPerBlock, UUID blockUUID) {
        PixelBlock.placedBlocks.add(this);
        this.uuid = blockUUID;

        this.pixelBlockLocation = originLocation.toBlockLocation();
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.pixelsPerBlock = pixelsPerBlock;
        this.owner = owner;

//        pixelBlockLocation.getBlock().setType(Material.GLASS);

        createWorld();
    }

    void createWorld() {
        File file = new File(plugin.getDataFolder().getPath()+ "/" +this.uuid.toString());

        if(!file.exists() || !file.isDirectory()) {
            System.out.println("NEUE WELT WURDE ERSTELLT");

            final WorldCreator worldCreator = new WorldCreator(plugin.getDataFolder().getPath() + "/" + this.uuid);
            worldCreator.type(WorldType.FLAT);
            worldCreator.generator(new EmptyChunkGenerator());
            World newWorld = Bukkit.createWorld(worldCreator);

            assert newWorld != null;
            newWorld.setGameRule(GameRule.DO_DAYLIGHT_CYCLE, false);

            Location borderStartLocation = newWorld.getSpawnLocation().clone().subtract(1, 1, 1);
            Location grassStartLocation = borderStartLocation.clone().subtract(worldGrassBorderWidth, 0, worldGrassBorderWidth);

            Bukkit.getScheduler().runTask(plugin, () -> {
                for (int x = 0; x < 18 + 2 * worldGrassBorderWidth; x++) {
                    for (int z = 0; z < 18 + 2 * worldGrassBorderWidth; z++) {
                        grassStartLocation.clone().add(x, 0, z).getBlock().setType(Material.GRASS_BLOCK);
                    }
                }
                for (int x = 0; x < 18 + 2 * worldGrassBorderWidth; x++) {
                    for (int z = 0; z < 18 + 2 * worldGrassBorderWidth; z++) {
                        grassStartLocation.clone().add(x, -1, z).getBlock().setType(Material.DIRT);
                    }
                }
                for (int x = 0; x < 18; x++) {
                    for (int z = 0; z < 18; z++) {
                        Location currentLocation = borderStartLocation.clone().add(x, 0, z);
                        if (currentLocation.x() == borderStartLocation.x() || currentLocation.z() == borderStartLocation.z()) {
                            currentLocation.getBlock().setType(Material.RED_CONCRETE);
                        } else if (currentLocation.x() == borderStartLocation.x() + 17 || currentLocation.z() == borderStartLocation.z() + 17) {
                            currentLocation.getBlock().setType(Material.RED_CONCRETE);
                        }
                    }
                }
                borderStartLocation.getBlock().setType(Material.WHITE_CONCRETE);
            });
        }
    }

    public void update() {
        Bukkit.getScheduler().runTask(plugin, () -> {
            for(Pixel pixel : this.pixels) { pixel.remove(); }
            pixels.clear();

            for (int x = 0; x < 16; x++) {
                for (int y = 0; y < 16; y++) {
                    for (int z = 0; z < 16; z++) {
                        Location relativeLocation = new Location(pixelBlockLocation.getWorld(), x, y, z);
                        Location blockLocation = Objects.requireNonNull(Bukkit
                                .getWorld(plugin.getDataFolder().getPath() + "/" + this.uuid))
                                .getSpawnLocation()
                                .clone()
                                .subtract(0, 1, 0)
                                .add(relativeLocation.x(), relativeLocation.y(), relativeLocation.z());
                        BlockData block = blockLocation.getBlock().getBlockData();

                        if(block.getMaterial() != Material.AIR) {
                            Pixel newPixel = new Pixel(relativeLocation, block, (1/pixelsPerBlock)-0.0001, 0.00005*pixelsPerBlock);
                            pixels.add(newPixel);
                        }
                    }
                }
            }

            for(Pixel pixel : this.pixels) {
                pixel.spawn(this.pixelBlockLocation);
            }
        });
    }

    public void saveToDB() {
        String url = "jdbc:sqlite:pixelblocks.db";
        List<UUID> uuids = new ArrayList<>();
        try (var conn = DriverManager.getConnection(url);
             var stmt = conn.createStatement()) {
            ResultSet pixelBlocksResult = stmt.executeQuery("SELECT * FROM pixelblocks;");
            while (pixelBlocksResult.next()) {
                uuids.add(UUID.fromString(pixelBlocksResult.getString("uuid")));
            }
        } catch (SQLException e) {
            System.err.println(e.getMessage());
        }

        if(!uuids.contains(this.uuid)) {
            try (var conn = DriverManager.getConnection(url)) {
                PreparedStatement prep = conn.prepareStatement(
                    "INSERT INTO pixelblocks(uuid, owner, locationWorldName, locationX, locationY, locationZ)" +
                        " VALUES(?, ?, ?, ?, ?, ?);"
                );
                prep.setString(1, this.uuid.toString());
                prep.setString(2, this.owner.toString());
                prep.setString(3, this.pixelBlockLocation.getWorld().getName());
                prep.setDouble(4, this.pixelBlockLocation.getX());
                prep.setDouble(5, this.pixelBlockLocation.getY());
                prep.setDouble(6, this.pixelBlockLocation.getZ());

                prep.executeUpdate();
            } catch (SQLException e) {
                System.err.println(e.getMessage());
            }
        }
    }

    public void removeFromDB() {
        Bukkit.getScheduler().runTask(plugin, () -> {
            String url = "jdbc:sqlite:pixelblocks.db";
            try (var conn = DriverManager.getConnection(url)) {
                PreparedStatement prep = conn.prepareStatement("DELETE FROM pixelblocks WHERE uuid = ?");
                prep.setString(1, this.uuid.toString());
                prep.executeUpdate();

                conn.close();
                prep.close();
            } catch (SQLException e) {
                System.err.println(e.getMessage());
            }
        });
    }

    public void place(Location placeLocation) {
        Bukkit.getScheduler().runTask(plugin, () -> {
            hitbox = (Interaction) pixelBlockLocation.getWorld().spawnEntity(
                    pixelBlockLocation.clone().add(0.5, -hitboxOffset, 0.5),
                    EntityType.INTERACTION
            );
            hitbox.setInteractionHeight(1F + 2*hitboxOffset);
            hitbox.setInteractionWidth(1F + 2*hitboxOffset);
            this.saveToDB();
            update();
        });
//        this.updateTaskID = Bukkit.getScheduler().scheduleSyncRepeatingTask(plugin, this::update, 100L, 100L);
    }

    public void remove() {
        this.removeFromDB();

        this.pixels.forEach(Pixel::remove);
        hitbox.remove();
//        pixelBlockLocation.getBlock().setType(Material.AIR);
        placedBlocks.remove(this);
    }

    public void delete() {
        this.remove();

        Bukkit.unloadWorld(plugin.getDataFolder().getPath() + "/" + this.uuid, true);
        try {
            FileUtils.deleteDirectory(plugin.getDataFolder().getPath() + "/" + this.uuid);
        } catch (IOException e) {
            System.err.println("World could not be deleted!");
        }
    }
}
